@@ -31,7 +31,7 @@ use XAO::Objects;
 use base XAO::Objects->load(objname => 'FS::Glue');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: Collection.pm,v 1.2 2002/01/04 01:47:37 am Exp $ =~ /(\d+\.\d+)/);
+($VERSION)=(q$Id: Collection.pm,v 1.7 2003/10/29 22:07:30 am Exp $ =~ /(\d+\.\d+)/);
 
 ###############################################################################
 
@@ -61,6 +61,28 @@ sub delete () {
 
 ###############################################################################
 
+=item describe () {
+
+Describes itself, returns a hash reference with at least the following
+elements:
+
+ type       => 'collection'
+ class      => class name
+ key        => key name
+
+=cut
+
+sub describe ($;$) {
+    my $self=shift;
+    return {
+        type    => @_ ? 'hash' : 'collection',
+        class   => $$self->{class_name},
+        key     => $$self->{key_name},
+    };
+}
+
+###############################################################################
+
 =item detach ()
 
 Not implemented, but safe to use.
@@ -97,6 +119,9 @@ method to retrieve multiple Hash references at once.
 If an object does not exist an error will be thrown, use exists() method
 to check if you really need to.
 
+Note: It does not check if the object still exists in the database! If
+you need to be sure that the object does exist use
+
 =cut
 
 sub get ($$) {
@@ -105,13 +130,17 @@ sub get ($$) {
     $self->throw("get - at least one ID required") unless @_;
 
     my @results=map {
-        $_ || $self->throw("get - no object ID given");
-        XAO::Objects->new(objname => $$self->{class_name},
-                          glue => $self->_glue,
-                          unique_id => $_,
-                          key_name => $$self->{key_name},
-                          list_base_name => $$self->{base_name},
-                         );
+        $_ ||
+            throw $self "get - no object ID given";
+        ref($_) &&
+            throw $self "get - should be a scalar, not a ".ref($_)." reference";
+        XAO::Objects->new(
+            objname         => $$self->{class_name},
+            glue            => $self->_glue,
+            unique_id       => $_,
+            key_name        => $$self->{key_name},
+            list_base_name  => $$self->{base_name},
+        );
     } @_;
 
     @_==1 ? $results[0] : @results;
@@ -119,6 +148,31 @@ sub get ($$) {
 
 ###############################################################################
 
+=item get_new ()
+
+Convenience method that returns new empty detached object of the type
+that collection operates on.
+
+=cut
+
+sub get_new ($) {
+    my $self=shift;
+    $self->glue->new(objname => $$self->{class_name});
+}
+
+###############################################################################
+
+=item glue ()
+
+Returns the Glue object which was used to retrieve the current object
+from.
+
+=cut
+
+# Implemented in Glue
+
+###############################################################################
+
 =item keys ()
 
 Returns unsorted list of all keys for all objects stored in that list.
