@@ -50,7 +50,7 @@ use XAO::Objects;
 use base XAO::Objects->load(objname => 'Atom');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: Glue.pm,v 1.22 2002/10/29 09:23:58 am Exp $ =~ /(\d+\.\d+)/);
+($VERSION)=(q$Id: Glue.pm,v 1.35 2003/11/13 00:54:00 am Exp $ =~ /(\d+\.\d+)/);
 
 ###############################################################################
 
@@ -146,13 +146,13 @@ sub new ($%) {
             $args->{empty_database} eq 'confirm' ||
                 throw $self "new - request for 'empty_database' is not 'confirm'ed";
 
-            $driver->initialize_database();
+            $driver->initialize_database;
         }
 
         ##
         # Loading data layout
         #
-        $$self->{classes}=$driver->load_structure();
+        $$self->{classes}=$driver->load_structure;
     }
 
     else {
@@ -253,11 +253,12 @@ Rough equivalent of:
 
 =cut
 
-sub destroy ($) {
-    my $self=shift;
+sub destroy ($;$) {
+    my ($self,$lists_only)=@_;
     foreach my $key ($self->keys) {
         my $type=$self->describe($key)->{type};
         next if $type eq 'key';
+        next if $lists_only && $type ne 'list';
         $self->delete($key);
     }
 }
@@ -297,34 +298,89 @@ object (formerly /Global):
 
  my $global=$odb->fetch('/');
 
+B<Experimental extension>: The following full forms can also be used:
+
+ xaofs://uri/Pages/P123/Hits/H234/name
+ xaofs://collection/class/Data::Page/5678/Hits/H234/name
+
+Both should yield the same result providing that 5678 is a Collection
+(see L<XAO::DS::FS::Collection>) code for P123.
+
 =cut
 
 sub fetch ($$) {
     my $self=shift;
     my $path=shift;
 
-    ##
-    # Normalizing and checking path
-    #
-    $path=$self->normalize_path($path);
-    substr($path,0,1) eq '/' || throw $self "fetch - bad path ($path)";
-
-    ##
-    # Going through the path and retrieving every element. Could be a
-    # little bit more optimal, but not much..
-    #
-    my $node=XAO::Objects->new(objname => 'FS::Global',
-                               glue => $self,
-                               uri => '/');
-    foreach my $nodename (split(/\/+/,$path)) {
-        next unless length($nodename);
-        if(ref($node)) {
-            $node=$node->get($nodename);
-        } else {
-            return undef;
-        }
+    $path =~ /^(xaofs:\/\/(.*?))?(\/.*)$/;
+    my $type;
+    my $full_path;
+    if($1) {
+        $type=$2;
+        $full_path=$path;
+        $path=$3;
     }
-    $node;
+
+    if(!defined $type || $type eq 'uri') {
+
+        ##
+        # Normalizing and checking path
+        #
+        $path=$self->normalize_path($path);
+        substr($path,0,1) eq '/' || throw $self "fetch - bad path ($path)";
+
+        ##
+        # Going through the path and retrieving every element. Could be a
+        # little bit more optimal, but not much..
+        #
+        my $node=XAO::Objects->new(objname => 'FS::Global',
+                                   glue => $self,
+                                   uri => '/');
+        foreach my $nodename (split(/\/+/,$path)) {
+            next unless length($nodename);
+            if(ref($node)) {
+                $node=$node->get($nodename);
+            } else {
+                return undef;
+            }
+        }
+
+        return $node;
+    }
+    elsif($type eq 'collection') {
+        $path =~ qr|^/class/(\w+(::\w+)*)(/.*)?$| ||
+            throw $self "fetch - wrong collection uri ($full_path)";
+
+        my $class=$1;
+        $path=$3;
+
+        my $node=$self->collection(class => $class);
+        if($path && $path ne '/') {
+            foreach my $nodename (split(/\/+/,$path)) {
+                next unless length($nodename);
+                if(ref($node)) {
+                    $node=$node->get($nodename);
+                } else {
+                    return undef;
+                }
+            }
+        }
+
+        return $node;
+    }
+}
+
+###############################################################################
+
+=item glue ()
+
+Returns reference to the Glue object the current object is received from.
+
+=cut
+
+sub glue ($) {
+    my $self=shift;
+    return $self->_glue;
 }
 
 ###############################################################################
@@ -354,6 +410,137 @@ sub objtype ($) {
 
 ###############################################################################
 
+=item reset ()
+
+Useful to bring glue to a usable state after some unknown software used
+it. If there is an active transaction -- it will be rolled back, if
+there are locked tables -- they will be unlocked.
+
+=cut
+
+sub reset ($) {
+    my $self=shift;
+
+    if($self->transact_active) {
+        $self->transact_rollback;
+    }
+
+    $self->_driver->reset;
+}
+
+###############################################################################
+
+=item transact_active ()
+
+Checks if a there is an active transaction at the moment. Can be used to
+avoid starting another one as transactions can't be nested.
+
+Example:
+
+  $odb->transact_begin unless $odb->transact_active;
+
+=cut
+
+sub transact_active ($) {
+    my $self=shift;
+    return $self->_driver->tr_ext_active;
+}
+
+###############################################################################
+
+=item transact_begin ()
+
+Begins new transaction. If transactions are not supported by the
+underlying driver it does nothing, see transact_can() method.
+
+If there is already an active transaction an error will be thrown and no
+new transaction will be started. Transactions can't be nested and
+each transact_begin() must be matched by a transact_commit() or
+transact_rollback().
+
+Automatic transact_rollback() is performed on destroying Glue or
+disconnecting. It is not done automatically on thrown errors, but if an
+error is never caught the transaction will be rolled back automatically
+at the program termination stage when Glue is destroyed. Beware of
+global error catching blocks and do proper cleanup if you use them.
+
+Example:
+
+  $odb->transact_begin;
+  try {
+      $order->put(order_total => 123.45);
+      $order->get('Products')->put($product_obj);
+      $odb->transact_commit;
+  }
+  otherwise {
+      my $e=shift;
+      $odb->transact_rollback;
+      $e->throw;
+  };
+
+=cut
+
+sub transact_begin ($) {
+    my $self=shift;
+    return $self->_driver->tr_ext_begin;
+}
+
+###############################################################################
+
+=item transact_can ()
+
+Returns boolean true if underlying driver supports transactions of false
+otherwise.
+
+Example:
+
+  $odb->transact_can ||
+      throw XAO::E::Sample "Transactions are not supported";
+
+=cut
+
+sub transact_can ($) {
+    my $self=shift;
+    return $self->_driver->tr_ext_can;
+}
+
+###############################################################################
+
+=item transact_commit ()
+
+Commits changes made during the current transaction. If there is no
+current transaction it will throw an error assuming that there was an
+out-of-sync transact_commit() or transact_rollback() somewhere before.
+
+If transactions are not supported the method will do nothing.
+
+=cut
+
+sub transact_commit ($) {
+    my $self=shift;
+    return $self->_driver->tr_ext_commit;
+}
+
+###############################################################################
+
+=item transact_rollback ()
+
+Rolls back (cancels) changes made during the current transaction. If
+there is no current transaction it will throw an error assuming that
+there was an out-of-sync transact_commit() or transact_rollback()
+somewhere before.
+
+If transactions are not supported the method will do nothing.
+
+=cut
+
+sub transact_rollback ($) {
+    my $self=shift;
+    return $self->_driver->tr_ext_rollback;
+}
+
+###############################################################################
+
 =item unlink ($)
 
 Alias to delete(), which is defined in derived classes - List and Hash.
@@ -405,7 +592,11 @@ Returns list of values for either Hash or List.
 
 sub values ($) {
     my $self=shift;
-    $self->get($self->keys());
+    my @k=$self->keys;
+    my $num=scalar(@k);
+    eprint ref($self)."::values - more then 100 keys ($num), consider scanning instead"
+        if $num > 100;
+    return $self->get(@k);
 }
 
 ###############################################################################
@@ -575,11 +766,7 @@ sub _field_default ($$) {
         $default='';
     }
     elsif($type eq 'integer' || $type eq 'real') {
-        if(!defined($desc->{minvalue}) && !defined($desc->{maxvalue})) {
-
-            # This happens only for 'real' type where we do not set any
-            # borders unless specified.
-            #
+        if(!defined $desc->{minvalue}) {
             $default=0;
         }
         elsif($desc->{minvalue} <= 0 &&
@@ -702,13 +889,13 @@ sub _retrieve_data_fields ($@) {
 ###############################################################################
 
 ##
-# Stores single value into hash object. Works only on Hash objects.
+# Stores multiple pre-checked for validity key/value pairs into hash
+# object. Works only on Hash objects.
 #
-sub _store_data_field ($$$$$) {
-    my $self=shift;
-    my ($name,$value)=@_;
+sub _store_data_fields ($$) {
+    my ($self,$data)=@_;
     my $table=$self->_class_description->{table};
-    $self->_driver->update_field($table,$$self->{unique_id},$name,$value);
+    $self->_driver->update_fields($table,$$self->{unique_id},$data);
 }
 
 ###############################################################################
@@ -876,7 +1063,6 @@ following structure, not a string:
  values       => array of values to be substituted into the SQL query
  classes      => hash with all classes and their aliases
  fields_list  => list of all fiels names
- tables_list  => list of all tables, their aliases and class names
  fields_map   => hash with the map of 'condition field name' => 'sql name'
  distinct     => list of fields to be unique
  order_by     => list of fields to sort on
@@ -899,6 +1085,7 @@ sub _build_search_query ($%) {
     # name into classes.
     #
     my $condition=$args->{conditions};
+    ### dprint "CONDITION: ",Dumper($condition);
     my %classes;
     my @values;
     my %fields_map;
@@ -915,8 +1102,7 @@ sub _build_search_query ($%) {
         $clause='';
         my $class_name=$$self->{class_name} ||
             $self->throw("_build_search_query - no 'class_name', not a List or Collection?");
-        $classes{index}='a';
-        $classes{$class_name}=$classes{index}++;
+        $self->_build_search_field_class(\%classes,$class_name,1,"",1);
     }
 
     ##
@@ -926,6 +1112,7 @@ sub _build_search_query ($%) {
     my @distinct;
     my @orderby;
     my $options=$args->{options};
+    my $debug;
     if($options) {
         foreach my $option (keys %$options) {
             if(lc($option) eq 'distinct') {
@@ -968,6 +1155,17 @@ sub _build_search_query ($%) {
             elsif(lc($option) eq 'limit') {
                 # pass through, handled in the driver
             }
+            elsif(lc($option) eq 'index') {
+                my $fn=$options->{$option};
+                my ($sqlfn,$fdesc,$class_name,$class_tag)=
+                    $self->_build_search_field(\%classes,$fn);
+                $classes{center_class}=$class_name;
+                $classes{center_tag}=$class_tag;
+                $classes{center_weight}=1000;
+            }
+            elsif(lc($option) eq 'debug') {
+                $debug=$options->{$option};
+            }
             else {
                 $self->throw("_build_search_query - unknown option '$option'");
             }
@@ -982,79 +1180,127 @@ sub _build_search_query ($%) {
         @return_fields{CORE::values %fields_map}=CORE::values %fields_map;
     }
 
+    ##
+    # Adding key name into fields we need.
+    #
+    my $key=$args->{key} ||
+        $self->throw("_build_search_query - no 'key' given");
+    my ($key_field)=$self->_build_search_field(\%classes,$key);
+
+    ##
+    # Removing unused top of the tree
+    #
+    my $top_class=$classes{top_class};
+    my $top_tag=$classes{top_tag};
+    my $up=$classes{up};
+    while(1) {
+        last if $up->{$top_class}->{$top_tag}->{name};
+        my $count=0;
+        my $ntc;
+        my $ntt;
+        foreach my $cin (keys %$up) {
+            foreach my $tin (keys %{$up->{$cin}}) {
+                if($up->{$cin}->{$tin}->{class} eq $top_class) {
+                    if(!$count) {
+                        $ntc=$cin;
+                        $ntt=$tin;
+                    }
+                    $count++;
+                }
+            }
+            last if $count>1;
+        }
+        last if $count>1;
+        delete $up->{$top_class};
+        $top_class=$classes{top_class}=$ntc;
+        $top_tag=$classes{top_tag}=$ntt;
+        $up->{$top_class}->{$top_tag}->{class}='';
+    }
+
+    ##
+    # Adding names to tables that are left in the tree but are not
+    # referred by any fields and therefore did not yet get their names.
+    #
+    foreach my $cin (keys %$up) {
+        foreach my $tin (keys %{$up->{$cin}}) {
+            if(!$up->{$cin}->{$tin}->{name}) {
+                my $ci=$classes{index}++;
+                $up->{$cin}->{$tin}->{name}=$ci;
+                $classes{names}->{$ci}={
+                    class       => $cin,
+                    tag         => $tin,
+                    weight      => 0,
+                    table       => $self->_class_description($cin)->{table},
+                };
+            }
+        }
+    }
+
     ##
     # Translating classes into table names and adding glue to join
     # tables together into the clause.
     #
-    my @tables_list;
-    delete $classes{index};
-    my $previous;
-    my $wrapped;
+    # We link up tables that are above our center point and then link
+    # down the rest. This takes care of all tables in correct order. SQL
+    # engine can of course reverse or alter the order, but generally
+    # this should be pretty optimal way of joining.
+    #
+    my ($cc,$ct)=@classes{'center_class','center_tag'};
+    my $wrapped=$clause && substr($clause,0,1) eq '(';
     my $glue=$self->_glue;
-    my %rev_classes;
-    foreach my $cn (keys %classes) {
-        my $ci=$classes{$cn};
-        if(ref($ci)) {
-            delete $ci->{max};
-            foreach my $i (CORE::values %$ci) {
-                $rev_classes{$i}=$cn;
-            }
+    while(1) {
+        my ($uc,$ut,$cn)=@{$up->{$cc}->{$ct}}{'class','tag','name'};
+        last unless $uc;
+        if(!$wrapped) {
+            $clause="($clause)" if $clause;
+            $wrapped=1;
         }
-        else {
-            $rev_classes{$ci}=$cn;
+        my $conn=$glue->_connector_name($cc,$uc);
+        if($conn) {
+            $conn=$self->_driver->mangle_field_name($conn);
+            my $un=$up->{$uc}->{$ut}->{name};
+            $clause.=" AND " if $clause;
+            $clause.="$un.unique_id=$cn.$conn";
         }
+        $up->{$cc}->{$ct}->{done}=1;
+        $cc=$uc;
+        $ct=$ut;
     }
-    foreach my $ci (sort keys %rev_classes) {
-        my $cn=$rev_classes{$ci};
-        my $desc=$self->_class_description($cn);
-        my $table=$desc->{table};
-        my $item={ table => $table,
-                   class => $cn,
-                   index => $ci,
-                 };
-        push(@tables_list,$item);
-
-        if($previous) {
+    foreach my $cc (keys %$up) {
+        foreach my $ct (keys %{$up->{$cc}}) {
+            my ($uc,$ut,$cn,$d)=@{$up->{$cc}->{$ct}}{'class','tag','name','done'};
+            next unless $uc && !$d;
             if(!$wrapped) {
                 $clause="($clause)" if $clause;
                 $wrapped=1;
             }
 
-            my $upper_class=$glue->upper_class($cn);
-            my $conn=$glue->_connector_name($cn,$upper_class);
-            if($conn) {
-                $conn=$ci . '.' . $self->_driver->mangle_field_name($conn);
 
+            my $conn=$glue->_connector_name($cc,$uc);
+            if($conn) {
+                $conn=$self->_driver->mangle_field_name($conn);
+                my $un=$up->{$uc}->{$ut}->{name};
                 $clause.=" AND " if $clause;
-                $clause.="$classes{$upper_class}.unique_id=$conn";
+                $clause.="$cn.$conn=$un.unique_id";
             }
         }
-        $previous=1;
     }
 
     ##
-    # Adding key name into fields we need. Forming an array of field
-    # names with key guaranteed to be the first.
+    # Moving key to the first position in the list of fields
     #
-    my $key=$args->{key} || $self->throw("_build_search_query - no 'key' given");
-    if($key ne 'unique_id') {
-        ($key)=$self->_build_search_field(\%classes,$key);
-    }
-    else {
-        my $class_alias=$tables_list[0]->{index};
-        $key=$class_alias . '.' . $key;
-    }
-    delete $return_fields{$key};
-    my @fields_list=($key, keys %return_fields);
+    delete $return_fields{$key_field};
+    my @fields_list=($key_field, keys %return_fields);
     undef %return_fields;
 
     ##
     # Composing SQL query out of data we have.
     #
+    my $c_names=$classes{names};
     my $sql='SELECT ';
     $sql.=join(',',@fields_list) .
           ' FROM ' .
-          join(',',map { $_->{table} . ' AS ' . $_->{index} } @tables_list);
+          join(',',map { $c_names->{$_}->{table} . ' AS ' . $_ } keys %{$c_names});
     $sql.=' WHERE ' . $clause if $clause;
 
     ##
@@ -1068,7 +1314,7 @@ sub _build_search_query ($%) {
     if(@distinct) {
         $sql.=' GROUP BY ' . join(',',@distinct);
     }
-    elsif(@fields_list>1 || @tables_list>1) {
+    elsif(@fields_list>1 || scalar(keys %$c_names)>1) {
         $sql.=' GROUP BY ' . join(',',$fields_list[0]);
     }
 
@@ -1084,7 +1330,10 @@ sub _build_search_query ($%) {
         }
     }
 
-    ### dprint "SQL: $sql";
+    ##
+    # To aid in debugging..
+    #
+    print STDERR "SEARCH SQL: $sql\n" if $debug;
 
     ##
     # Returning resulting hash
@@ -1096,7 +1345,6 @@ sub _build_search_query ($%) {
         classes => \%classes,
         fields_list => \@fields_list,
         fields_map => \%fields_map,
-        tables_list => \@tables_list,
         distinct => \@distinct,
         order_by => \@orderby,
         post_process => $post_process,
@@ -1120,125 +1368,178 @@ sub _build_search_field ($$$) {
     my $self=shift;
     my $classes=shift;
     my $lha=shift;
+    my $glue=$self->_glue;
 
-    my $class_name=$$self->{class_name} ||
-        $self->throw("_build_search_field - no 'class_name', not a List or Collection?");
+    my $up=$classes->{up};
+    $up=$classes->{up}={} unless $up;
 
-    $classes->{$class_name}=$classes->{index}++ unless $classes->{$class_name};
+    ##
+    # Optimizing stupid things like 'D/../E' into 'E'
+    #
+    while($lha=~/^(.*\/)?\w.*?\/\.\.\/(.*)$/) {
+        $lha=(defined($1) ? $1 : '') . $2;
+    }
 
-    my $class_index;
-
-    my $desc=$$self->{class_description};
+    ##
+    # Splitting field name into parts if it looks like path either
+    # absolute or relative. Real field name is the last element, popping
+    # it back into $lha.
+    #
     my @path=split(/\/+/,$lha);
     $lha=pop @path;
+
+    my $class_name;
+    my $class_tag=1;
+    if(@path && $path[0] eq '') {
+        $class_name='FS::Global';
+        shift @path;
+    }
+    else {
+        $class_name=$$self->{class_name} ||
+            $self->throw("_build_search_field - no 'class_name', not a List or Collection?");
+        while(@path && $path[0] eq '..') {
+            $class_name=$glue->upper_class($class_name) ||
+                throw $self "_build_search_field - no upper class for $class_name";
+            shift @path;
+        }
+    }
+
+    $self->_build_search_field_class($classes,$class_name,$class_tag,"",1);
+
+    my $class_desc=$self->_class_description($class_name);
+
     for(my $i=0; $i!=@path; $i++) {
         my $n=$path[$i];
 
-        my $fd=$desc->{fields}->{$n} ||
+        my $fd=$class_desc->{fields}->{$n} ||
             $self->throw("_build_search_field - unknown field '$n' in $lha");
         $fd->{type} eq 'list' ||
             $self->throw("_build_search_field - '$n' is not a list in $lha");
-        $class_name=$fd->{class};
-        $desc=$self->_class_description($class_name);
+        my $n_class=$fd->{class};
 
-        my $tag=$i+1==@path ? '' : $path[$i+1];
-        if($tag eq '*' || $tag=~/^\d+$/) {
+        my $n_tag=$i+1==@path ? '' : $path[$i+1];
+        if($n_tag eq '*') {
             $i++;
-
-            ##
-            # XXX - To fix this we need a lot of changes in the code, we
-            # need no single name based table dependency, but some sort
-            # of dependencies stack for each final field. Would be nice
-            # to have though!
-            #
-            $i+1==@path ||
-                throw $self "_build_search_field - modifiers supported on last element only !TODO! (".join('/',@path,$lha).")";
+            $classes->{tag_index}||='ta';
+            $n_tag=$classes->{tag_index}++;
+        }
+        elsif($n_tag=~/^\d+$/) {
+            $i++;
+        }
+        else {
+            $n_tag=1;
         }
 
-        if($tag eq '') {
-            if(exists $classes->{$class_name}) {
-                $class_index=$classes->{$class_name};
-                if(ref($class_index)) {
-                    $class_index=$class_index->{999} ||
-                                 $class_index->{$class_index->{max}} ||
-                        throw $self "_build_search_field - OOPS, can't find class index ($class_name,$lha)";
-                }
-            }
-            else {
-                $classes->{$class_name}=$classes->{index}++;
-            }
-        }
-        elsif($tag eq '*') {
-            $class_index=$classes->{index}++;
-            if(exists $classes->{$class_name}) {
-                my $d=$classes->{$class_name};
-                if(ref($d)) {
-                    my $m=$d->{max}+1;
-                    $d->{$m}=$class_index;
-                    $d->{max}=$m;
-                }
-                else {
-                    $d={
-                        max => 1000,
-                        999 => $d,
-                        1000 => $class_index,
-                    };
-                }
-                $classes->{$class_name}=$d;
-            }
-            else {
-                $classes->{$class_name}={
-                    max => 1000,
-                    1000 => $class_index,
-                };
-            }
-        }
-        elsif($tag =~ /^\d+$/) {
-            if(exists $classes->{$class_name}) {
-                my $d=$classes->{$class_name};
-                if(ref($d)) {
-                    if(exists $d->{$tag}) {
-                        $class_index=$d->{tag};
-                    }
-                    else {
-                        $class_index=$classes->{index}++;
-                        $d->{$tag}=$class_index;
-                        $d->{max}=$tag if $tag>$d->{max};
-                    }
-                }
-                else {
-                    $class_index=$classes->{index}++;
-                    $d={
-                        $tag    => $class_index,
-                        999     => $d,
-                        max     => $tag>999 ? $tag : 999,
-                    };
-                }
-                $classes->{$class_name}=$d;
-            }
-            else {
-                $class_index=$classes->{index}++;
-                $classes->{$class_name}={
-                    max => $tag,
-                    $tag => $class_index,
-                };
-            }
-        }
+        $self->_build_search_field_class($classes,$n_class,$n_tag,
+                                                  $class_name,$class_tag);
+        $class_name=$n_class;
+        $class_tag=$n_tag;
+        $class_desc=$self->_class_description($class_name);
     }
 
-    my $field_desc=$desc->{fields}->{$lha} ||
-        $self->throw("_build_search_field - unknown field '$lha'");
-
+    my $class_index=$up->{$class_name}->{$class_tag}->{name};
     if(!$class_index) {
-        my $ci=$classes->{$class_name};
-        $class_index=ref($ci) ? ($ci->{999} || $ci->{$ci->{max}}) : $ci;
-        $class_index ||
-            throw $self "_build_search_field - OOPS, can't get class_index ($class_name, $lha)";
+        $classes->{index}||='a';
+        $class_index=$classes->{index}++;
+        $up->{$class_name}->{$class_tag}->{name}=$class_index;
+        $classes->{names}->{$class_index}={
+            class       => $class_name,
+            tag         => $class_tag,
+            weight      => 0,
+            table       => $class_desc->{table},
+        };
     }
 
-    $lha=$class_index . '.' . $self->_driver->mangle_field_name($lha);
+    ##
+    # Special condition for 'unique_id' field names
+    #
+    my $field_desc=$lha eq 'unique_id' ? {} : $class_desc->{fields}->{$lha};
+    $field_desc || $self->throw("_build_search_field - unknown field '$lha'");
 
-    ($lha,$field_desc);
+    ##
+    # Counting number of fields using that table, index have more weight
+    # and unique index even more. If not overriden in options that is
+    # going to be out center table.
+    #
+    my $inc=$field_desc->{unique} ? 20 : ($field_desc->{index} ? 10 : 1);
+    my $weight=$classes->{names}->{$class_index}->{weight}+=$inc;
+    if(!$classes->{center_weight} || $classes->{center_weight}<$weight) {
+        $classes->{center_weight}=$weight;
+        $classes->{center_class}=$class_name;
+        $classes->{center_tag}=$class_tag;
+    }
+
+    ##
+    # We don't need to mangle field name if it's a unique_id field
+    #
+    if($lha eq 'unique_id') {
+        $lha=$class_index . '.' . $lha;
+    }
+    else {
+        $lha=$class_index . '.' . $self->_driver->mangle_field_name($lha);
+    }
+
+    return ($lha,$field_desc,$class_name,$class_tag);
+}
+
+###############################################################################
+
+sub _build_search_field_class ($$$$$$) {
+    my ($self,$classes,$n_class,$n_tag,$p_class,$p_tag)=@_;
+
+    my $up=$classes->{up};
+
+    if(!$p_class && (!$up->{$n_class} || !$up->{$n_class}->{$n_tag})) {
+        if(!$classes->{top_class}) {
+            $classes->{top_class}=$n_class;
+            $classes->{top_tag}=$n_tag;
+        }
+        elsif($classes->{top_class} ne $n_class) {
+            my $uc=$classes->{top_class};
+            while($uc ne $n_class) {
+                $uc=$self->_glue->upper_class($uc);
+                last unless $uc;
+                $up->{$uc}->{1}={
+                    class   => '',
+                    tag     => 1,
+                };
+                my ($ctc,$ctt)=@{$classes}{'top_class','top_tag'};
+                @{$up->{$ctc}->{$ctt}}{'class','tag'}=($uc,1);
+                @{$classes}{'top_class','top_tag'}=($uc,$uc eq $n_class ? $n_tag : 1);
+            }
+            if(!$uc) {
+                $uc=$n_class;
+                my $ut=$n_tag;
+                while(1) {
+                    if(!$up->{$uc} || !$up->{$uc}->{$ut}) {
+                        $up->{$uc}->{$ut}={
+                            class   => $self->_glue->upper_class($uc),
+                            tag     => 1,
+                        };
+                        $uc=$up->{$uc}->{$ut}->{class};
+                        $ut=1;
+                    }
+                    last if $up->{$uc}->{$ut};
+                }
+            }
+        }
+        elsif($classes->{top_tag} ne $n_tag) {
+            my $uc=$classes->{top_class};
+            $uc=$self->_glue->upper_class($uc) ||
+                throw $self "_build_search_field_class - no upper class for $uc";
+            $up->{$uc}->{1}={
+                class   => '',
+                tag     => 1,
+            };
+        }
+    }
+
+    return if $up->{$n_class} && $up->{$n_class}->{$n_tag};
+
+    $up->{$n_class}->{$n_tag}={
+        class => $p_class,
+        tag   => $p_tag,
+    };
 }
 
 ###############################################################################
@@ -1258,11 +1559,6 @@ sub _build_search_clause ($$$$$$) {
     my $post_process=shift;
     my $condition=shift;
 
-    ##
-    # Initial index for classes is 'a'
-    #
-    $classes->{index}='a' unless $classes->{index};
-
     ##
     # Checking if the condition has exactly three elements
     #
@@ -1305,7 +1601,6 @@ sub _build_search_clause ($$$$$$) {
                                             $post_process,
                                             $rha);
 
-
         my $clause;
         if($op eq 'or' || $op eq '||') {
             $clause="($lhv OR $rhv)";
@@ -1365,6 +1660,9 @@ sub _build_search_clause ($$$$$$) {
     elsif($op eq 'cs') {
         $clause="$field LIKE '" . '%' . $rha_escaped . '%' . "'";
     }
+    elsif($op eq 'sw') {
+        $clause="$field LIKE '" . $rha_escaped . '%' . "'";
+    }
     elsif($op eq 'ws') {
         my $tf;
         ($clause,$tf)=$self->_driver->search_clause_ws($field,$rha,$rha_escaped);
@@ -1388,10 +1686,10 @@ sub _build_search_clause ($$$$$$) {
         }
     }
     else {
-        $self->throw("_build_search_clause - unknown operator '$op'");
+        throw $self "_build_search_clause - unknown operator '$op'";
     }
 
-    $clause;
+    return $clause;
 }
 
 ###############################################################################
@@ -1415,6 +1713,7 @@ sub _list_setup ($) {
 
     my $kdesc=$$self->{class_description}->{fields}->{$$self->{key_name}};
     $$self->{key_format}=$kdesc->{key_format};
+    $$self->{key_length}=$kdesc->{key_length} || 30;
     $$self->{key_unique_id}=$kdesc->{key_unique_id};
 }
 
@@ -1444,7 +1743,15 @@ sub _list_unlink_object ($$$) {
     my $object=$self->get($name) || $self->throw("_list_unlink_object - no object exists (name=$name)");
     my $class_desc=$object->_class_description();
 
-    $object->destroy();
+    ##
+    # Asking destroy to delete lists only, otherwise it will delete each
+    # and every field separately thus being very very slow.
+    #
+    $object->destroy(1);
+
+    ##
+    # And now dropping the row itself.
+    #
     $self->_driver->delete_row($class_desc->{table},
                                $$object->{unique_id});
 }
@@ -1583,7 +1890,9 @@ sub _add_data_placeholder ($%) {
     }
     elsif ($type eq 'integer') {
         $fdesc{minvalue}=-0x80000000 unless defined($fdesc{minvalue});
-        $fdesc{maxvalue}=0x7FFFFFFF unless defined($fdesc{maxvalue});
+        if(!defined($fdesc{maxvalue})) {
+            $fdesc{maxvalue}=$fdesc{minvalue}<0 ? 0x7FFFFFFF : 0xFFFFFFFF;
+        }
         $driver->add_field_integer($table,$name,$fdesc{index},$fdesc{unique},
                                    $fdesc{minvalue},$fdesc{maxvalue},$fdesc{default},$connected);
     }
@@ -1645,6 +1954,10 @@ sub _add_list_placeholder ($%) {
         throw $self "_add_list_placeholder - key_format must include either <\$RANDOM\$> or <\$AUTOINC\$> ($key_format)";
     }
 
+    my $key_length=$args->{key_length} || 30;
+    $key_length < 255 ||
+        throw $self "_add_list_placeholder - key_length ($key_length) must be less then 255";
+
     XAO::Objects->load(objname => $class);
 
     my $table=$args->{table};
@@ -1667,22 +1980,23 @@ sub _add_list_placeholder ($%) {
                 throw $self "_add_list_placeholder - such table ($table) is already used";
         }
 
-        $driver->add_table($table,$key,$connector);
+        $driver->add_table($table,$key,$key_length,$connector);
 
         $$glue->{classes}->{$class}={
             table => $table,
             fields => {
                 $key => {
-                    type => 'key',
-                    refers => $self->objname,
-                    key_format => $key_format,
+                    type        => 'key',
+                    refers      => $self->objname,
+                    key_format  => $key_format,
+                    key_length  => $key_length,
                 }
             }
         };
         if(defined($connector)) {
             $$glue->{classes}->{$class}->{fields}->{$connector}={
-                type => 'connector',
-                refers => $self->objname,
+                type        => 'connector',
+                refers      => $self->objname,
             }
         }
 
@@ -1701,6 +2015,7 @@ sub _add_list_placeholder ($%) {
                          refers     => $self->objname,
                          key_format => $key_format,
                          key_seq    => 1,
+                         maxlength  => $key_length,
                        });
     $driver->store_row('Global_Fields',
                        'field_name',$connector,
@@ -1764,19 +2079,18 @@ sub _drop_data_placeholder ($$) {
 # Instead of dropping each object individually we just drop entire
 # tables here. Potentially very dangerous.
 #
-sub _drop_list_placeholder ($$;$) {
-    my $self=shift;
-    my $name=shift;
-    my $recursive=shift;
+sub _drop_list_placeholder ($$;$$) {
+    my ($self,$name,$recursive,$upper_class)=@_;
 
     my $desc=$recursive || $self->_field_description($name);
     my $class=$desc->{class};
     my $glue=$self->_glue;
     my $cdesc=$$glue->{classes}->{$class};
     my $cf=$cdesc->{fields};
+
     foreach my $fname (keys %{$cf}) {
         if($cf->{$fname}->{type} eq 'list') {
-            $self->_drop_list_placeholder($fname,$cf->{$fname});
+            $self->_drop_list_placeholder($fname,$cf->{$fname},$class);
         }
     }
 
@@ -1806,7 +2120,13 @@ sub _drop_list_placeholder ($$;$) {
     delete $$glue->{classes}->{$class};
     delete $$glue->{list_keys_cache}->{$class};
     delete $$glue->{connectors_cache}->{$class};
-    delete $$glue->{classes}->{$self->objname}->{fields}->{$name};
+
+    if($recursive) {
+        delete $$glue->{classes}->{$upper_class}->{fields}->{$name};
+    }
+    else {
+        delete $$glue->{classes}->{$self->objname}->{fields}->{$name};
+    }
 
     $self->_driver->drop_table($table);
 }
